]> git.sesse.net Git - cubemap/blobdiff - server.h
Use C++11 std::mutex and std::lock_guard instead of our RAII wrapper.
[cubemap] / server.h
index a3e032d88f71a2f66b9613781d64053a6f32b143..68e72c4d543165e938a3e63810395727629c85ae 100644 (file)
--- a/server.h
+++ b/server.h
@@ -1,29 +1,34 @@
 #ifndef _SERVER_H
 #define _SERVER_H 1
 
-#include <pthread.h>
 #include <stddef.h>
 #include <stdint.h>
 #include <sys/epoll.h>
 #include <sys/types.h>
 #include <time.h>
 #include <map>
+#include <memory>
+#include <mutex>
+#include <queue>
 #include <string>
 #include <vector>
 
+#include "tlse.h"
+
 #include "client.h"
 #include "stream.h"
 #include "thread.h"
 
+class Acceptor;
 class ClientProto;
 struct Stream;
 
 #define EPOLL_MAX_EVENTS 8192
 #define EPOLL_TIMEOUT_MS 20
 #define MAX_CLIENT_REQUEST 16384
+#define REQUEST_READ_TIMEOUT_SEC 60
 
 class CubemapStateProto;
-class MarkPool;
 class StreamProto;
 
 class Server : public Thread {
@@ -35,61 +40,88 @@ public:
        std::vector<ClientStats> get_client_stats() const;
 
        // Set header (both HTTP header and any stream headers) for the given stream.
-       void set_header(const std::string &stream_id,
+       void set_header(int stream_index,
                        const std::string &http_header,
                        const std::string &stream_header);
 
// Set that the given stream should use the given mark pool from now on.
        // NOTE: This should be set before any clients are connected!
-       void set_mark_pool(const std::string &stream_id, MarkPool *mark_pool);
+       void set_pacing_rate(int stream_index, uint32_t pacing_rate);
 
        // These will be deferred until the next time an iteration in do_work() happens,
        // and the order between them are undefined.
        // XXX: header should ideally be ordered with respect to data.
-       void add_client_deferred(int sock);
-       void add_data_deferred(const std::string &stream_id, const char *data, size_t bytes);
+       void add_client_deferred(int sock, Acceptor *acceptor);
+       void add_data_deferred(int stream_index, const char *data, size_t bytes, uint16_t metacube_flags);
 
        // These should not be called while running, since that would violate
        // threading assumptions (ie., that epoll is only called from one thread
        // at the same time).
        CubemapStateProto serialize();
        void add_client_from_serialized(const ClientProto &client);
-       void add_stream(const std::string &stream_id, size_t bytes_received, Stream::Encoding encoding);
-       void add_stream_from_serialized(const StreamProto &stream, int data_fd);
-       void set_backlog_size(const std::string &stream_id, size_t new_size);
-       void set_encoding(const std::string &stream_id, Stream::Encoding encoding);
+       int add_stream(const std::string &url, size_t bytes_received, size_t prebuffering_bytes, Stream::Encoding encoding, Stream::Encoding src_encoding);
+       int add_stream_from_serialized(const StreamProto &stream, int data_fd);
+       int lookup_stream_by_url(const std::string &url) const;
+       void set_backlog_size(int stream_index, size_t new_size);
+       void set_prebuffering_bytes(int stream_index, size_t new_amount);
+       void set_encoding(int stream_index, Stream::Encoding encoding);
+       void set_src_encoding(int stream_index, Stream::Encoding encoding);
+       void add_gen204(const std::string &url, const std::string &allow_origin);
+       void create_tls_context_for_acceptor(const Acceptor *acceptor);
 
 private:
-       // Mutex protecting queued_data only. Note that if you want to hold both this
-       // and <mutex> below, you will need to take <mutex> before this one.
-       mutable pthread_mutex_t queued_data_mutex;
+       // Mutex protecting queued_add_clients.
+       // Note that if you want to hold both this and <mu> below,
+       // you will need to take <mu> before this one.
+       mutable std::mutex queued_clients_mutex;
 
        // Deferred commands that should be run from the do_work() thread as soon as possible.
        // We defer these for two reasons:
        //
        //  - We only want to fiddle with epoll from one thread at any given time,
        //    and doing add_client() from the acceptor thread would violate that.
-       //  - We don't want the input thread(s) hanging on <mutex> when doing
-       //    add_data(), since they want to do add_data() rather often, and <mutex>
+       //  - We don't want the input thread(s) hanging on <mu> when doing
+       //    add_data(), since they want to do add_data() rather often, and <mu>
        //    can be taken a lot of the time.
        //      
-       // Protected by <queued_data_mutex>.
-       std::vector<int> queued_add_clients;
-       std::map<std::string, std::string> queued_data;
+       // Protected by <queued_clients_mutex>.
+       std::vector<std::pair<int, Acceptor *>> queued_add_clients;
 
        // All variables below this line are protected by the mutex.
-       mutable pthread_mutex_t mutex;
+       mutable std::mutex mu;
+
+       // All streams.
+       std::vector<std::unique_ptr<Stream>> streams;
+
+       // Map from URL to index into <streams>.
+       std::map<std::string, int> stream_url_map;
 
-       // Map from stream ID to stream.
-       std::map<std::string, Stream *> streams;
+       // Map from URL to CORS Allow-Origin header (or empty string).
+       std::map<std::string, std::string> ping_url_map;
 
        // Map from file descriptor to client.
        std::map<int, Client> clients;
 
+       // A list of all clients, ordered by the time they connected (first element),
+       // and their file descriptor (second element). It is ordered by connection time
+       // (and thus also by read timeout time) so that we can read clients from the
+       // start and stop processing once we get to one that isn't ready to be
+       // timed out yet (which means we only have to look at each client exactly once,
+       // save for the first element of the queue, which is always checked).
+       //
+       // Note that when we delete a client, we don't update this queue.
+       // This means that when reading it, we need to check if the client it
+       // describes is still exists (ie., that the fd still exists, and that
+       // the timespec matches).
+       std::queue<std::pair<timespec, int>> clients_ordered_by_connect_time;
+
        // Used for epoll implementation (obviously).
        int epoll_fd;
        epoll_event events[EPOLL_MAX_EVENTS];
 
+       // For each TLS-enabled acceptor, our private server context for its key pair.
+       std::map<const Acceptor *, TLSContext *> tls_server_contexts;
+
        // The actual worker thread.
        virtual void do_work();
 
@@ -108,10 +140,29 @@ private:
        // but it's cheaper than taking it in and out all the time.
        void process_client(Client *client);
 
+       // If the TLS library wants to write anything to this client,
+       // output it. Returns true if the processing should go to sleep
+       // (an error, or lack of outgoing buffer space).
+       bool send_pending_tls_data(Client *client);
+
+       // Reads regular data fro ma socket. Returns -1 if the processing
+       // should go to sleep (an error, or no data available yet), otherwise
+       // the number of bytes read.
+       int read_nontls_data(Client *client, char *buf, size_t max_size);
+
+       // Reads (decrypted) data from a TLS socket. Returns -1 if the processing
+       // should go to sleep (an error, or no data available yet), otherwise
+       // the number of bytes read. The buffer will be used as scratch space
+       // for TLS data, so it can be overwritten by more bytes than what is returned.
+       int read_tls_data(Client *client, char *buf, size_t max_size);
+
        // Close a given client socket, and clean up after it.
        void close_client(Client *client);
 
-       // Parse the HTTP request. Returns a HTTP status code (200/400/404).
+       // Listen for a different set of epoll events.
+       void change_epoll_events(Client *client, uint32_t events);
+
+       // Parse the HTTP request. Returns a HTTP status code (200/204/400/404).
        int parse_request(Client *client);
 
        // Construct the HTTP header, and set the client into
@@ -119,15 +170,16 @@ private:
        void construct_header(Client *client);
 
        // Construct a generic error with the given line, and set the client into
-       // the SENDING_ERROR state.
+       // the SENDING_SHORT_RESPONSE state.
        void construct_error(Client *client, int error_code);
 
-       // TODO: This function should probably die.
-       Stream *find_stream(const std::string &stream_id);
+       // Construct a 204, and set the client into the SENDING_SHORT_RESPONSE state.
+       void construct_204(Client *client);
 
        void process_queued_data();
+       void skip_lost_data(Client *client);
 
-       void add_client(int sock);
+       void add_client(int sock, Acceptor *acceptor);
 };
 
 #endif  // !defined(_SERVER_H)
@@ -35,61 +40,88 @@ public:
                        const std::string &stream_header);
 
-       // Set that the given stream should use the given mark pool from now on.
+       // Set that the given stream should use the given max pacing rate from now on.
        // NOTE: This should be set before any clients are connected!